@@ -4,7 +4,7 @@ BlockchainProtocol.py: contains the BlockchainProtocol class.
 """
 
 
-import time, traceback, collections
+import time, traceback, collections, logging
 from twisted.internet import reactor
 from twisted.protocols.basic import LineReceiver
 
@@ -14,7 +14,8 @@ from pybc.util import string2bytes, bytes2string
 class BlockchainProtocol(LineReceiver):
     """
     This is a Twisted protocol that exchanges blocks with the peer at the other
-    end, and floods valid non-block messages. Used by both servers and clients.
+    end, and floods valid non-block messages (transactons). Also lets state
+    components be requested. Used by both servers and clients.
     
     It is a line-oriented protocol with occasional blocks of raw data.
     
@@ -70,10 +71,17 @@ class BlockchainProtocol(LineReceiver):
         
         """
         
-        print "Made a connection!"
+        # Let the peer know we are connected. It will tell us if we should keep
+        # the connection or not.
+        keep = self.factory.peer.made_connection(self)
         
-        # Let the peer know we are connected.
-        self.factory.peer.made_connection(self)
+        if not keep:
+            # We alread have too many incoming connections. Drop the connection.
+            logging.debug("Dropping unwanted connection")
+            self.disconnect()
+            return
+            
+        logging.info("Made a connection!")
         
         if self.incoming:
             # They connected to us. Send a greeting, saying what network we're
@@ -97,7 +105,7 @@ class BlockchainProtocol(LineReceiver):
         # We're no longer a connection that the peer should send things to.
         self.factory.peer.lost_connection(self)
         
-        print "Protocol lost connection. Reason: ", reason
+        logging.info("Protocol lost connection. Reason: {}".format(reason))
         
         
         if self.inv_watcher is not None and self.inv_watcher.active():
@@ -114,6 +122,9 @@ class BlockchainProtocol(LineReceiver):
         """
         
         reactor.callFromThread(self.transport.loseConnection)
+        
+        # TODO: Do something to stop us trying to send messages over this
+        # connection.
     
     def send_message(self, parts):
         """
@@ -211,6 +222,9 @@ class BlockchainProtocol(LineReceiver):
                 # Send them ADDR messages, one per known peer.
                 
                 for host, port, time_seen in self.factory.peer.get_peers():
+                    if time_seen is None:
+                        # This is a bootstrap peer
+                        continue
                     # Send the peer's host and port in an ADDR message
                     self.send_message(["ADDR", host, port, time_seen])
             elif parts[0] == "ADDR":
@@ -240,7 +254,7 @@ class BlockchainProtocol(LineReceiver):
                 # Decode the hash they have
                 block_hash = string2bytes(parts[1])
                 
-                if (not self.factory.peer.blockchain.has_block(block_hash) and 
+                if (self.factory.peer.blockchain.needs_block(block_hash) and 
                     block_hash not in self.inv_queued and
                     block_hash not in self.block_queued):
                     
@@ -265,8 +279,8 @@ class BlockchainProtocol(LineReceiver):
                     # because messages queue.
                     self.send_message(["BLOCK", bytes2string(block.to_bytes())])
                 else:
-                    print "Can't send missing block: '{}'".format(
-                        bytes2string(block_hash))
+                    logging.error("Can't send missing block: '{}'".format(
+                        bytes2string(block_hash)))
             
             elif parts[0] == "BLOCK":
                 # They have sent us a block. Add it if it is valid.
@@ -308,25 +322,25 @@ class BlockchainProtocol(LineReceiver):
                         # We have it (still). Send them the block transaction.
                         self.send_message(["TX", bytes2string(transaction)])
                     else:
-                        print "Lost transaction: '{}'".format(
-                            bytes2string(transaction_hash))  
+                        logging.error("Lost transaction: '{}'".format(
+                            bytes2string(transaction_hash)))
                 else:
-                    print "Can't send missing transaction: '{}'".format(
-                        bytes2string(transaction_hash))
+                    logging.error("Can't send missing transaction: '{}'".format(
+                        bytes2string(transaction_hash)))
             elif parts[0] == "TX":
                 # They have sent us a transaction. Add it to our blockchain.
                 
                 # Decode the transaction bytes
                 transaction_bytes = string2bytes(parts[1])
                 
-                print "Incoming transaction."
+                logging.debug("Incoming transaction.")
                 
                 if self.factory.peer.blockchain.transaction_valid_for_relay(
                     transaction_bytes):
                     
                     # This is a legal transaction to accept from a peer (not
                     # something like a block reward).
-                    print "Transaction acceptable from peer."
+                    logging.info("Transaction acceptable from peer.")
                 
                     # Give it to the blockchain as bytes. The blockchain can
                     # determine whether to forward it on or not and call the
@@ -334,16 +348,72 @@ class BlockchainProtocol(LineReceiver):
                     # (True or False).
                     self.factory.peer.send_transaction(
                         transaction_bytes)
+                        
+            elif parts[0] == "GETSTATE":
+                # They're asking us for a StateComponent from our State, with
+                # the given hash. We just serve these like a dumb server.
+                
+                # We assume this is under our current State. If it's not, we'll
+                # return a nort found message and make them start over again
+                # from the top.
+                
+                # What StateComponent do they want?
+                state_hash = string2bytes(parts[1])
+                
+                # Go get it
+                state_component = \
+                    self.factory.peer.blockchain.state.get_component(state_hash)
+                    
+                if state_component is None:
+                    # Complain we don't have that. They need to start over from
+                    # the state hash in the latest block.
+                    logging.warning("Peer requested nonexistent StateComponent "
+                        "{} vs. root hash {}".format(parts[1], 
+                        bytes2string(
+                        self.factory.peer.blockchain.state.get_hash())))
+                    self.send_message(["NOSTATE", bytes2string(state_hash)])
+                else:
+                
+                    logging.debug("Fulfilling request for StateComponent "
+                        "{}".format(parts[1]))
+                    
+                    # Pack up the StateComponent's data as bytes and send it
+                    # along. They'll know which one it was when they hash it.
+                    self.send_message(["STATE", 
+                        bytes2string(state_component.data)])
+                        
+            elif parts[0] == "NOSTATE": 
+                # They're saying they don't have a StateComponent we asked for.
+                logging.warning("Peer says they do not have StateComponent: "
+                    "{}".format(parts[1]))
+                    
+            elif parts[0] == "STATE":
+                # They're sending us a StateComponent we probably asked for.
+                
+                # Unpack the data
+                component_bytestring = string2bytes(parts[1])
+            
+                # Give it to the Blockchain. It knows how to handle these
+                # things. TODO: Maybe put a queue here, since this potentially
+                # does the whole state rebuilding operation.
+                self.factory.peer.blockchain.add_state_component(
+                    component_bytestring)
+                    
+                # Tell the peer to request more StateComponents.
+                # TODO: This is going to blow up.
+                self.factory.peer.request_state_components()
                   
             elif parts[0] == "ERROR":
                 # The remote peer didn't like something.
                 # Print debugging output.
-                print "Error from remote peer: {}".format(" ".join(parts[1:]))
+                logging.error("Error from remote peer: {}".format(" ".join(
+                    parts[1:])))
             
             else:
                 # They're trying to send a command we don't know about.
                 # Complain.
-                print "Remote host tried unknown command {}".format(parts[1])
+                logging.error("Remote host tried unknown command {}".format(
+                    parts[1]))
                 self.send_message(["ERROR", parts[0]])
             
             if not self.incoming:    
@@ -353,8 +423,8 @@ class BlockchainProtocol(LineReceiver):
                     self.remote_address[1], int(time.time()))
             
         except:
-            print "Exception processing command: {}".format(parts)
-            traceback.print_exc()
+            logging.error("Exception processing command: {}".format(parts))
+            logging.error(traceback.format_exc())
             
             # Disconnect from people who send us garbage
             self.disconnect()
@@ -387,11 +457,11 @@ class BlockchainProtocol(LineReceiver):
                 block_hash = self.inv_queue.popleft()
                 self.inv_queued.remove(block_hash)
                 
-                if (not self.factory.peer.blockchain.has_block(block_hash) and
+                if (self.factory.peer.blockchain.needs_block(block_hash) and
                     block_hash not in self.block_queued):
-                    # We don't have the block in our blockchain or in our queue
-                    # to add to the blockchain, so get it.
-                    # TODO: Check to see if we requested it recently.
+                    # We need this block and don't have it in our queue to add
+                    # to the blockchain, so get it. TODO: Check to see if we
+                    # requested it recently.
                     
                     self.send_message(["GETDATA", bytes2string(block_hash)])
                 
