@@ -3,13 +3,14 @@ Peer.py: contains the Peer class.
 
 """
 
-import random, time, threading
+import random, time, threading, logging
 
 from twisted.internet import reactor, endpoints
 
 from pybc.ClientFactory import ClientFactory
 from pybc.ServerFactory import ServerFactory
 from pybc.util import time2string, bytes2string
+import pybc.science
 
 import sqliteshelf
 
@@ -22,7 +23,7 @@ class Peer(object):
     
     def __init__(self, network, version, blockchain, peer_file = ":memory:", 
         external_address=None, port=8007, optimal_connections=10, 
-        tick_period=60, peer_timeout = 60 * 60 * 48):
+        connections_per_batch=5, tick_period=60, peer_timeout = 60 * 60 * 48):
         """
         Make a new Peer in the given network (identified by a string), with the
         given version integer, that keeps its blocks in the given Blockchain.
@@ -46,6 +47,8 @@ class Peer(object):
         optimal_connections is the number of connections we want to have. We
         will periodically try to get new connections or drop old ones.
         
+        TODO: Actually drop connections if we have too many.
+        
         tick_period is how often to tick and ping our peers/broadcast our
         address/try to make new connections/drop extra connections, in seconds.
         
@@ -74,6 +77,9 @@ class Peer(object):
         # Remember our optimal number of connections
         self.optimal_connections = optimal_connections
         
+        # And how many we should make per batch
+        self.connections_per_batch = connections_per_batch
+        
         # Remember our peer remembering timeout
         self.peer_timeout = peer_timeout
         
@@ -83,9 +89,6 @@ class Peer(object):
         # Make a Twisted ServerFactory
         self.server = ServerFactory(self)
         
-        # Listen with it
-        self.endpoint.listen(self.server)
-        
         # Make a Twisted ClientFactory to make outgoing connections
         self.client = ClientFactory(self)
         
@@ -102,8 +105,6 @@ class Peer(object):
         
         # Remember our tick frequency
         self.tick_period = tick_period
-        # And schedule a tick right away.
-        reactor.callLater(0, self.tick)
         
         # Make a lock so this whole thing can basically be a monitor
         self.lock = threading.RLock()
@@ -111,6 +112,25 @@ class Peer(object):
         # Remember if we need to repoll our peers for blocks
         self.repoll = False
         
+        # Keep a set of recently requested StateComponents, so we don't ask for
+        # the same one over and over again.
+        # TODO: No other queueing stuff is in Peer. Should this be?
+        self.recently_requested_state_components = set()
+        
+        # Remember our total tick count
+        self.tick_count = 0
+        
+        # A little bit after we get the last block in a string of blocks, check
+        # to see if we need StateComponents. This holds the Twisted callLater
+        # call ID.
+        self.state_download_timer = None
+        
+        # Listen with the endpoint, so we can get connections
+        self.endpoint.listen(self.server)
+        
+        # Schedule a tick right away.
+        reactor.callLater(0, self.tick)
+        
     def connect(self, host, port):
         """
         Make a new connection to the given host and port, as a client.
@@ -121,7 +141,7 @@ class Peer(object):
             # Remember that we are connecting to this host
             self.outgoing_hosts.add(host)
             
-            print "Connecting to {} port {}".format(host, port)
+            logging.info("Connecting to {} port {}".format(host, port))
             
             # Queue making a connection with our ClientFactory in the Twisted
             # reactor.
@@ -133,11 +153,21 @@ class Peer(object):
         Called when a connection is made. Add the connection to the set of
         connections.
         
+        Returns True if the connection was added to the set and is to be kept,
+        or False if we alread have too many connections and want to drop this
+        one.
+        
         """
         
         with self.lock:
-            # Add the connection to our list of connections
-            self.connections.add(connection)
+            if (len(self.connections) > 2 * self.optimal_connections and 
+                connection.incoming):
+                # Don't accept this incoming connection, we have too many.
+                return False
+            else:
+                # Add the connection to our list of connections
+                self.connections.add(connection)
+                return True
         
     def lost_connection(self, connection):
         """
@@ -182,6 +212,10 @@ class Peer(object):
         
         with self.lock:
         
+            if time_seen > int(time.time()) + 10 * 60:
+                # Don't accept peers from more than 10 minutes into the future.
+                return
+        
             if self.known_peers.has_key(host):
                 # We know of this peer, but we perhaps ought to update the last
                 # heard from time
@@ -190,9 +224,9 @@ class Peer(object):
                 # them
                 known_port, our_time_seen = self.known_peers[host]
                 
-                if time_seen > our_time_seen and time_seen <= time.time():
-                    # They heard from them more recently, but not in the future.
-                    # Update our time last seen.
+                if (our_time_seen is None or time_seen > our_time_seen):
+                    # They heard from them more recently. Update our time last
+                    # seen.
                     self.known_peers[host] = (known_port, time_seen)
             else:
                 # This is a new peer.
@@ -214,74 +248,121 @@ class Peer(object):
         """
         See if we have the optimal number of outgoing connections. If we have
         too few, add some.
-        
-        TODO: Broadcast our address.
-        
         """
         
         with self.lock:
         
             # How many connections do we have right now?
             current_connections = len(self.connections)
-            print "Tick from localhost port {}: {} of {} connections".format(
-                self.port, current_connections, self.optimal_connections)
+            logging.info("Tick {} from localhost port {}: {} of {} "
+                "connections".format(self.tick_count, self.port, 
+                current_connections, self.optimal_connections))
             
             for connection in self.connections:
-                print "\tTo {} port {}".format(*connection.remote_address)
+                logging.info("\tTo {} port {}".format(
+                    *connection.remote_address))
             
-            print "{} outgoing connections:".format(len(self.outgoing_hosts))
+            logging.info("{} outgoing connections:".format(len(
+                self.outgoing_hosts)))
             for host in self.outgoing_hosts:
-                print "\t To {}".format(host)
+                logging.info("\t To {}".format(host))
             
-            print "Blockchain height: {}".format(
-                len(self.blockchain.hashes_by_height))
-            print "Blocks known: {}".format(
-                self.blockchain.get_block_count())
-            print "Blocks pending: {}".format(
-                len(self.blockchain.pending_callbacks))
-            print "Transactions pending: {}".format(len(
-                self.blockchain.transactions))
+            logging.info("Blockchain height: {}".format(
+                len(self.blockchain.hashes_by_height)))
+            logging.info("Blocks known: {}".format(
+                self.blockchain.get_block_count()))
+            logging.info("Blocks pending: {}".format(
+                len(self.blockchain.pending_callbacks)))
+            logging.info("State available for verification and mining:"
+                " {}".format(self.blockchain.state_available))
+            logging.info("Transactions pending: {}".format(len(
+                self.blockchain.transactions)))
+                
+            # Calculate the average block time in seconds for the last few
+            # blocks. It may be None, meaning we don't have enough history to
+            # work it out.
+            block_time = self.blockchain.get_average_block_time()
+            if block_time is not None:
+                logging.info("Average block time: {} seconds".format(
+                    block_time))
             
+            logging.info("Blockchain disk usage: {} bytes".format(
+                self.blockchain.get_disk_usage()))
+            
+            if self.tick_count % 60 == 0:
+                # Also log some science stats every 60 ticks (by default, 1 hour)
+                pybc.science.log_event("chain_height", 
+                    len(self.blockchain.hashes_by_height))
+                pybc.science.log_event("pending_transactions",
+                    len(self.blockchain.transactions))
+                pybc.science.log_event("connections", current_connections)
+                if block_time is not None:
+                    # We have an average block time, so record that.
+                    pybc.science.log_event("block_time", block_time)
+                
+                # We want to log how much space our databases (block store and
+                # state) take.
+                pybc.science.log_event("blockchain_usage", 
+                    self.blockchain.get_disk_usage())
+            
+            # Request whatever state components we need, egardless if whether we
+            # recently requested them, in case our download got stalled
+            self.recently_requested_state_components = set()
+            self.request_state_components()
+
+                    
             if (len(self.outgoing_hosts) < self.optimal_connections and 
                 len(self.known_peers) > 0):
                 # We don't have enough outgoing connections, but we do know some
                 # peers.
                 
-                for _ in xrange(self.optimal_connections - 
-                    len(self.outgoing_hosts)):
-                
-                    # For each connection we want but don't have
-                
-                    # Find a peer we aren't connected to and connect to them
-                    host = random.sample(self.known_peers, 1)[0]
+                for i in xrange(min(self.connections_per_batch, 
+                    self.optimal_connections - len(self.outgoing_hosts))):
+                    # Try several connections in a batch.
                     
-                    # Try at most 100 times to find a host we aren't connected
-                    # to
-                    attempt = 1
-                    while host in self.outgoing_hosts and attempt < 100:
-                        # Try a new host
+                    for _ in xrange(self.optimal_connections - 
+                        len(self.outgoing_hosts)):
+                    
+                        # For each connection we want but don't have
+                    
+                        # Find a peer we aren't connected to and connect to them
                         host = random.sample(self.known_peers, 1)[0]
                         
-                        # Increment attempt
-                        attempt += 1
+                        # Try at most 100 times to find a host we aren't
+                        # connected to
+                        attempt = 1
+                        while host in self.outgoing_hosts and attempt < 100:
+                            # Try a new host
+                            host = random.sample(self.known_peers, 1)[0]
+                            
+                            # Increment attempt
+                            attempt += 1
+                            
+                        # TODO: This always makes two attempts at least
+                            
+                        if attempt < 100:
+                            # We found one!
+                            # Connect to it.
+                            
+                            # Get the port (and discard the last heard from
+                            # time)
+                            port, _ = self.known_peers[host]
                         
-                    # TODO: This always makes two attempts at least
-                        
-                    if attempt < 100:
-                        # We found one!
-                        # Connect to it.
-                        
-                        # Get the port (and discard the last heard from time)
-                        port, _ = self.known_peers[host]
-                    
-                        # Connect to it.
-                        self.connect(host, port)
+                            # Connect to it.
+                            self.connect(host, port)
+                        else:
+                            # No more things we can try connecting to
+                            break
                     
             # Throw out peers that are too old. First compile a list of their
             # hostnames.
             too_old = []
 
             for host, (port, last_seen) in self.known_peers.iteritems():
+                if last_seen is None:
+                    # This is a bootstrap peer. Don't remove it.
+                    continue
+                    
                 if time.time() - last_seen > self.peer_timeout:
                     # We haven't heard from/about this node recently enough.
                     too_old.append(host)
@@ -291,10 +372,14 @@ class Peer(object):
                 del self.known_peers[host]
                 
             # Broadcast all our hosts.
-            print "{} known peers:".format(len(self.known_peers))
+            logging.info("{} known peers".format(len(self.known_peers)))
             for host, (port, last_seen) in self.known_peers.iteritems():
-                print "\tPeer {} port {} last seen {}".format(host, port, 
-                    time2string(last_seen))
+                if last_seen is None:
+                    # This is a bootstrap peer, so don't announce it.
+                    continue
+                    
+                logging.debug("\tPeer {} port {} last seen {}".format(host,
+                    port, time2string(last_seen)))
                 for connection in self.connections:
                     connection.send_message(["ADDR", host, port, last_seen])
             
@@ -307,18 +392,21 @@ class Peer(object):
             # Do we need to re-poll for blocks?
             if self.repoll:
                 self.repoll = False
-                print "Repolling due to unverifiable block."
+                logging.warning("Repolling due to unverifiable block.")
                 
                 # Compose just one message
                 message = (["GETBLOCKS"] + 
                     [bytes2string(block_hash) for block_hash in 
                     self.blockchain.get_block_locator()])
-                
                 for connection in self.connections:
                     connection.send_message(message) 
             
             
-            print "Saving..."
+            logging.info("Saving...")
+            # Keep track of how long this takes.
+            save_start = time.clock()
+            
+            pybc.science.start_timer("sync")
             
             # Sync the blockchain to disk. This needs to lock the blockchain, so
             # the blockchain must never wait on the peer while holding its lock.
@@ -329,12 +417,21 @@ class Peer(object):
             # Sync the known peers to disk
             self.known_peers.sync()
             
-            print "Saved to disk"
+            pybc.science.stop_timer("sync")
+            
+            # How long did it take?
+            save_end = time.clock()
+            
+            logging.info("Saved to disk in {:.2} seconds".format(save_end - 
+                save_start))
                         
+            # Count this tick as having happened
+            self.tick_count += 1
+                    
             # Tick again later
             reactor.callLater(self.tick_period, self.tick)
         
-            print "Tick complete."
+            logging.info("Tick complete.")
         
     def announce_block(self, block_hash):
         """
@@ -399,11 +496,24 @@ class Peer(object):
         """
         
         
-        if not self.blockchain.has_block(block.block_hash()):
-            # It's new.
+        if self.blockchain.needs_block(block.block_hash()):
+            # It's something the blockchain is interested in.
             # Queue the block, and, if valid, announce it
             self.blockchain.queue_block(block, 
                 callback=self.was_block_valid)
+        
+        if self.state_download_timer is not None:
+            try:
+                # We started a timer earlier. Cancel it if it hasn't run
+                self.state_download_timer.cancel()
+            except:
+                # It probably already happened
+                pass
+        
+        # In a little bit, if thre are no new blocks before then, download
+        # StateComponents that we need (if any)
+        self.state_download_timer = reactor.callLater(0.5,
+            self.request_state_components)
             
     def send_transaction(self, transaction):
         """ 
@@ -415,6 +525,41 @@ class Peer(object):
         # Add the transaction, and, if valid, announce it
         self.blockchain.add_transaction(transaction, 
             callback=self.was_transaction_valid)
+     
+    def request_state_components(self):
+        """
+        Ask all our peers for any StateComponents that our Blockchain needs, and
+        that we have not recently requested.
+        
+        """
+        
+        with self.lock:
+            if not self.blockchain.state_available:
+                # We need to ask for StateComponents from our peers until we get
+                # all of them.
+                for requested_component in \
+                    self.blockchain.get_requested_state_components():
+                    
+                    if (requested_component in 
+                        self.recently_requested_state_components):
+                        
+                        # We just asked for this one
+                        continue
+                    
+                    logging.debug("Requesting StateComponent {} from all "
+                        "peers".format(bytes2string(requested_component)))
+                    
+                    # Compose just one message per request
+                    message = ["GETSTATE", bytes2string(requested_component)]
+                    
+                    for connection in self.connections:
+                        # Send it to all peers. Hopefully one or more of them
+                        # will send us a StateComponent back.
+                        connection.send_message(message) 
+                        
+                    # Remember having asked for this component
+                    self.recently_requested_state_components.add(
+                        requested_component)
         
     def run(self):
         """
