@@ -2,7 +2,12 @@
 # coin.py: a coin implemented on top of pybc
 
 import struct, hashlib, traceback, time, threading, logging, itertools
-import pyelliptic
+
+try:
+    import pyelliptic
+except:
+    # pyelliptic didn't load. Either it's not installed or it can't find OpenSSL
+    import emergency_crypto_munitions as pyelliptic
 
 import pybc
 import transactions, util, sqliteshelf
